library randomlist;

import "dart:async";
import "dart:math";

part '../randomlist/randomlistitem.dart';
part '../randomlist/itempicker.dart';

/**
 *  A RandomList is a collection that lets client code 'pick' a member item that is returned at
 * random.
/// By default, each member item has an equal chance of being picked. However, different integer 'weights'
///  can be specified for each item to allow for uneven chances of being picked randomly.
///
/// By default, items in the list stay on the list--meaning they can be picked an infinite number
/// of times. Optionally, however, an integer representing the 'lifetime' of an item can be specified
/// for items that are meant to expire.

 * Items are maintained in the order added to the table to allow for lists that assign
 * some kind of meaning to the order, such as items at the beginning are good and those at the end are bad.
 *
 * A RandomList is ideal to use for something such as an encounter table.
 *
 *
 * TODO have a randomRemove func?  addAt()?  insertBefore( id )
 * should track the useCount - how many times item has been used - to prevent overuse
 *
 * // REV history
 *

 */

class RandomList {

  static const String VERSION = "1.0";

  // event types
  static const String ADDED = "added";
  //static const String REMOVED = "removed";
  //static const String EXPIRED = "expired";
  //static const String PICKED = "picked";
  //static const String PICKING = "picking";
  //static const String FAILED = "failed";


  StreamController<RandomListItem> _expiredCon = new StreamController<RandomListItem>.broadcast();
  StreamController<RandomListItem> pickingCon = new StreamController<RandomListItem>.broadcast();
  StreamController<RandomListItem> pickedCon = new StreamController<RandomListItem>.broadcast();
  StreamController<RandomListItem> _removedCon = new StreamController<RandomListItem>.broadcast();
  StreamController<RandomListItem> failedCon = new StreamController<RandomListItem>.broadcast();

  Stream get onExpired => _expiredCon.stream;
  Stream get onFailed => failedCon.stream;
  Stream get onRemoved => _removedCon.stream;
  Stream get onPicking => pickingCon.stream;
  Stream get onPicked => pickedCon.stream;


  /// by default drops 5% for a standard 100 weight
  static const int DEFAULT_WEIGHT_DROP_FACTOR = 5;

  int _totalWeight = 0;

  List<RandomListItem> get itemList => _itemList;

  // the list of RandomListItems, stored in order added
  List<RandomListItem> _itemList = new List<RandomListItem>();
  Map<String, RandomListItem> _itemMap = new Map<String, RandomListItem>();

  /**
   *  the most recent thing picked on the list
   */
  RandomListItem mostRecentItem;

  /**
   * If the choose can't find something suitable within this
   * number of tries, it gives up and returns a null item.
   */
  int pickAttemptLimit = 20;

  String id = "";

  /*
   *  the last list item that used up its lifetime and was removed
   */
  RandomListItem lastExpired;

  // should be read only
  num lastBaseRoll = 0;

  num lastModifiedRoll = 0;

  Random random = new Random();

  ItemPicker picker;

  /**
   * how many items in list
   */
  int get length => _itemList.length;

  //////////////////////////////////////////////////////////////////
  /**
   * constructor
   * Every list has an id so that it can be targeted if it is a child.
   */
  RandomList( this.id ) {
    picker = new ItemPicker( this );
  }



  /////////////////////////////////////////////////////////////////////

  /**
   * Add a new object onto the list. wraps it in a RandomListItem
  ///
  /// By default every item has the same chance (weight) of selection (100) and an infinite lifetime.
  ///
  /// @param id item id
  /// @param obj the arbitrary inner object to wrap in a RandomListItem - the simplest uses don't need an item and just work off of an id
  /// @param weight  default is 100   -- this can be overriden if a weightObject is specified
  /// @param weightDecrease default is 0  The amount that the item weight should be lowered every time it is picked. -1 (don't drop) by default -- weight drop is not compatible if a weightObject is specified
  /// @param weightIncrease every time something is picked from the list and it is not this item,
  ///      the weight of this item will rise by this amount not compatible with weight object.
  /// @param userData condition string - an arbitrary string that can be used by a filter to filter the item before a pick.
  /// *
  /// * wt drop / increase could have been a single # that is interpreted whether positive or negative.
  /// *
  /// *
  /// * @returns The RandomListItem created
   */
  RandomListItem add( String _id, {
        Object obj : null,
        int weight : 100,
        int lifetime : -1,
        int weightDecrease :0,
        int weightIncrease :0,
        bool removeOnZero:false,
        int maxWeight:-1,
        Object userData :null} ) {

    //trace("list add() id, obj, wt, life, wtdrop", _id, obj, wt, lifetime, weightDrop );
    //trace("add ", weightIncrease );

    //trace("#### ADD() creating new item:", _id, obj);
    //var bitem:RandomListItem = new RandomListItem( _id, obj, this, wt, weightDrop,weightIncrease, userData);
    RandomListItem item = new RandomListItem( _id.toUpperCase(), obj, this, weight, weightDecrease, weightIncrease, userData );
    item.removeOnZero = removeOnZero;
    if( maxWeight > -1) item.maxWeight = maxWeight;
    //trace("### END ADD");

    item.lifetime = lifetime;

    _itemList.add( item );
    _itemMap[ item.id.toUpperCase() ] = item;

    return item;
  }

  /**
   *  * add a pre existing RandomListItem to list
   *  NOTE: * only used for cloning and reconstituting from XML
   */
  ///
  ///
  addExistingItem( RandomListItem item ) {
    item.setParentList( this );
    _itemList.add( item );
    _itemMap[ item.id ] = item;
    //recalcTotal();   this will happen on pick right?
  }

  /**
   * Add an item directly to a named sub list
   * return NULL_LIST_ITEM on failure
   * @param objectId how to refer to this object
   * @param object - they payload, obj to wrap if any
   * @listId
   * @return
   *
   *
   *
   *
   * ??? can you ever recalc thresholds without recalcing wts? I don't think so
   */
  RandomListItem addToChildList(String  listId, String objectId, [Object object=null, int wt = 100, int lifetime = -1, int weightDrop = 0, int weightIncrease = 0, Object userData = null] ) {

    RandomListItem result = RandomListItem.NULL_LIST_ITEM;

    var item = getListItem( listId.toUpperCase() );
    if( item == null ) return result;
    else {
      // find the list
      var sublist = item.payload;
      return sublist.add( objectId, object, wt, lifetime, weightDrop, weightIncrease, userData );
    }
  }


  /**
     * called BEFORE itemHasBeenPicked()
     * if expired, dispatches EXPIRED event with item as data
     * @returns whether it expired or not
     */
    bool adjustLifetime( RandomListItem item ) {
      //print("adjusting lifetime ${item.id} ${item.lifetime}");

      if( item.lifetime == -1 ) {
        //trace("...ignoring, infinite life");
        return false;  // did NOT adjust, it has infinite life
      } else {

        // decrement life
        item.lifetime--;

        print("adjusting, new lifetime ${item.lifetime}");

        if( item.lifetime == 0 ) {  // it has expired

          item.parentList.removeItem( item.id );

          print("expiring: " + item.id);
          // expire; this is the last usage
          lastExpired = item;

          _expiredCon.add( item );  // alert stream listeners

          return true;
        } else return false;

      }
    }


    /**
     *
     * Change the weight of an item that does not have
     * a dynamic wt function
     *
     * @returns whether it changed
     */
    bool changeWeight(String id, int newWeight ) {

      RandomListItem item = getItem( id );

        if( item != null ) {

          item.weight = newWeight ;
          //item.parentList.recalcWeights();  // this happens on a pick
          return true;
        } else {
          return false;
        }
      }


  /**
   * remove all items
   */
  clear() =>  _itemList = new List<RandomListItem>();

  /**
   * deep copy this list
   */
  RandomList clone(String newId) {
    RandomList newList = new RandomList( newId );

    _itemList.forEach((item) {
       RandomListItem newItem = item.clone();
       newList.addExistingItem( newItem );
    });

    return newList;
  }

  /**
   * does this list contain the item either directly or on a child list?
   */
  bool contains(String itemId ) {
    bool ret = false;

    itemId = itemId.toUpperCase();

    // is it contained directly?
    if( _itemMap.containsKey(itemId) ) return true;
    //else print("not directly contained $itemId");

    // else go through all child lists and query them
    Iterable itemsThatAreLists = _itemList.where((item){
       return (item.payload is RandomList);
    });

    itemsThatAreLists.firstWhere((item){
      RandomList childList = item.payload as RandomList;
      //print("...examining list ${childList.id}");
      if(childList.contains(itemId) ) {
         ret = true;
         return true;
      } else {
         return false;
      }
    }, orElse:(){
      //print("no match");
    });

    return ret;
  }

  /**
    * Currently if weight drop below zero it wraps around to maxWeight
    */
   doWeightDecrease( RandomListItem item ) {
     //print("doWeightDecrease");
     adjust( int w ) {

         int itemWt = w;


         // drop weight if required
         int newWeight = itemWt - item.weightDropAmount;

         //print("dropping weight for ${item.id} to $newWeight");

         if( newWeight <= 0 ) {
                if( item.removeOnZero ) {
                  removeItem( item.id );
                } else {
                  newWeight = item.maxWeight;  // ?? wrap around ??
                  //trace("...new weight wrapped around = " + newWeight);
                  changeWeight( item.id, newWeight );
                }
          } else {
                // positive wt
                //trace("...changing weight");
                changeWeight( item.id, newWeight );
          }

     } // adjust


     if( (item != null) && item.weightDrops ){

       var w = item.weight;

       if( w is int ) {
         //print("int weight");
         adjust( w );
       } else if( w is Future) {
         //print("got a Future returned from wt.");
         w.then((wt){
           adjust(w);
         });
       } else {
         print("unknown item weight type");
       }

     } else {

       // if still on list and weight the same, then do nothing
     }
   }


   doWeightIncrease( RandomListItem itemPicked ) {
        //print("doWeightIncrease");
       // go through list looking for items that are not this item and that have
       // a weightIncrease greater than 0

       for (RandomListItem i in _itemList) {
         int inc = i.weightIncreaseAmount;

         if( itemPicked == i ) {
           //trace("skipping the picked item");
           continue;
         }

         if( inc > 0 ) {
           print("increasing wt of ${i.id} by $inc");
           i.addWeight( inc );
           //recalcWeights( cb );  this happens before each pick
         }
       }


     }


  RandomListItem getItem(String itemId ) {
    itemId = itemId.toUpperCase();

    // is it contained directly?
    if( _itemMap.containsKey(itemId) ) {
       return _itemMap[ itemId ];
    } else {

      // else go through all child lists and query them
      Iterable itemsThatAreLists = _itemList.where((item){
         return (item.payload is RandomList);
      });

      RandomListItem ret;

      // grab the first list that says it contains the item
      itemsThatAreLists.firstWhere((item){
        RandomList childList = item.payload as RandomList;
        //print("...examining list ${childList.id}");
        if(childList.contains(itemId) ) {
           ret = childList.getItem(itemId);
           return true;
        } else {
           return false;
        }
      }, orElse:(){
        // ignore
      });

      //if( ret != null) return
      return ret;
    }
  }



  /**
    * Get all items from this list and any child lists.
    */
   List<RandomListItem> getAllItems( ) {
     //print("$id - getAllItems()");

     List<RandomListItem> list = new List();

     _itemList.forEach((item){
       //print(item.id);
       list.add( item ); // always add the item itself
       if( item.payload is RandomList ) {
          list.addAll( item.payload.getAllItems() );  // if list add children
       }
     });

     //print("get all items ${list.length}");
     return list;
   }

   /**
    * get the current lifetime of the item specified
    */
   int getLifetime(String  itemId ) {
     return getListItem( itemId ).lifetime;
   }

   /**
      * Get the item containing the list object.
      * specify item id
      */
   RandomListItem getListItem(String listId ) {
       var has = contains( listId );
       if( ! has ) {
         //print("not have");
          return null;
       } else {
         // contains reported true
         RandomListItem item = getItem( listId );
         //print(item);
         if( item.payload is RandomList ) return item;
         else return null;
       }
     }

   /**
    * an internal notification in case subclass needs to do some custom response.
    */
   itemHasBeenPicked( RandomListItem item  ) {
     //trace("item has been picked");

     doWeightDecrease( item );

     doWeightIncrease( item );
   }


   /**
    * Randomly return a payload OBJECT contained in the list somewhere.
    *
    * If the item is itself a randomList, recursively choose from that one.
    *
    * first thing this does is to recalc the wts of the items -- to allow for dynamic weights
    * - based on runtime factors like vars
    *
    * decrements the lifetime and removes if necessary
    *
    * If item weight drops below zero it is reset to maximum weight.
    * The callback will get the returned object value as an arg.
    *
    * @param mod a # to add( subtract) to/from the selection roll
    * @param filter allows a filter object to be passed into to screen the pickedItems against before allowing a good choice


    * NOTE: forceId DOES NOT work at the moment
    * @param forceId  force the item with the specified Id to be returned from the list -- ignore filters
    *
    * @returns a future that will get the payload
    */
   Future pick( {num mod : 0, RandomListFilter filter:null, String forceId :null})  {
     print("$id beginning pick process");

     Completer completer = new Completer();

     // fire PICKING
     pickingCon.add( null );

     try {
         recalcWeights( ).then((v){

           picker.pick(completer, mod:mod, filter:filter, forceId:forceId);

         });
     } catch( err ) {
       print("RandomList.recalcWeights() error\n" + err.message);
     }

     return completer.future;
   } // pick



   /**
    * remove from the list and fire an event
    * ??? should do a recalc?
    */
   RandomListItem removeItem( String id ) {
     id = id.toUpperCase();

     var ret;

     // do I own it directly?
     RandomListItem item = _itemMap[ id ];
     if( item != null) {
       //print("removing directly");
       _itemMap.remove(id);
       _itemList.remove(item);
       ret = item;
       //print("adding to controller: $item");
       _removedCon.add(item);
     } else {
       //print("finding child list...");

       var parentItem = findParentListItem( id );
       if( parentItem != null ) {
         var list = parentItem.payload; // the list
         ret = list.removeItem( id );
         //print("adding to controller: $item");
         _removedCon.add(ret);
       }
     }

     return ret;
   }


   /**
    * gets the item that has a RandomList that
    * contains the child.
    */
   RandomListItem findParentListItem( String childId) {
       childId = childId.toUpperCase();

       var ret;

       // you can simply get the item...
       var item = getItem( childId );

       if( item != null ) {
           // get the id of the parent list
           var parentId = item.parentList.id;
           // and get THAT item now...
           ret = getItem( parentId );
       }

       return ret;
   }


   Future _waitForWeight( RandomListItem item ) {
     //print("waitForWeight ${item.id}");
     Completer completer = new Completer();

     var w = item.weight;

     if( w is int) completer.complete( w );
     else if(w is Future) {
         w.then((value) => completer.complete(value));
     } else throw new Exception("unknown weight type $w for ${item.id}");

     return completer.future;
   }


  /**
   *  want to process all async, then report done at end
   *
   *      * go through every item on the list and get its weight
    * its percent chance to be picked is the wt divided by total wt
   */
  Future recalcWeights(  ) {
    Completer completer = new Completer();

    //print("recalc weights()");
    // recalc Total weights;

    _totalWeight = 0;
    num _threshCount = 0;

    // once wts are calced, cache them
    List cachedWeightsList = new List();

    List<Future> waitingList = new List<Future>();

    // for each item call _waitForWeight
    // and add any futures to list
    Future.forEach(_itemList, (RandomListItem item) {
      // for each item, add it's waiting function to the list
      waitingList.add( _waitForWeight( item ) );

    }).then((f) {
      // once we have filled the waiting list
      // wait for all Futures to complete

      // gets a list of all values returned
      Future.wait(waitingList).then((valueList) {
         //print("done waiting ${valueList.length}");

         // populate the cachedWeightList
         for( var ndx = 0; ndx < _itemList.length; ndx++ ) {
           var item = _itemList[ ndx ];
           var value = valueList[ ndx ];
           if( value <= 0 && item.removeOnZero ) {
             // don't add anything, but remove it frorm list
             // can't modify list during iteration
             new Future(() {
                removeItem( item.id );
                print("removing on zero wt: " + item.id);
             }) ;

           } else {
             _totalWeight += value; // add to total

             cachedWeightsList.add([item, value ] );  // cache as temp List obj
             //print("list wt callback, adding $value" );

           }
         }
         //print("   total wt: $_totalWeight");

         // all weights calced
         // now calc thresholds

         //print("calc thresh...");

        cachedWeightsList.forEach((entry){
          //print("*** examine entry");
          num percent = entry[1] / _totalWeight;  // entry[1] is the cached wt
          var item = entry[0] as RandomListItem;
          item.percent = percent * 100;  // put it more in the range 0 - 100%

          _threshCount += percent;
          item.threshold = _threshCount;

          //print(item.id + " calc percent: " +  item.percent.toString());
          //print(item.id + "   " + item.threshold.toString() );

        });

        //print("finished thresh");
        completer.complete(); // all done

      });

    });

    return completer.future;
  }


   /**
    * Set the lifetime of the item specified to an arbitrary number.
    */
   setLifetime(String itemId, int lifetime ) => getItem( itemId ).lifetime = lifetime;


   String toString() {
     return "[RandomList: $id]\n${getAllItems().join(', ')}";
   }
} // RandomList


typedef RandomListFilterCallback(var item, bool acceptOrReject );

/**
 *
 * function to determine whether an item is a candidate to be returned from a random call
 *
 * @param the item to potentially filter.
 * @param attemptsRemaining is there to try to mitigate too tough constraints -
 * so calling function can see if list is having hard time coming up with something that passes the filter.
 * @param The callback will get a boolean as an arg, indicating whether to allow
 * the item to be picked or not.
 */
typedef void RandomListFilter( RandomListItem item, int attemptsRemaining, RandomListFilterCallback callback);





