part of randomlist;


typedef void WeightAccepter(int weight);

typedef void WeightFunction( RandomListItem item, Object weightData, WeightAccepter accepter );

/**
 * Represents an item that can be randomly served up by a RandomList.
 *
 * Very often it's useful the item carries a 'payload,' around an arbitrary object
 * of another type, e.g. a game entity, etc.
 * RandomLists themselves can be list items, allowing you
 * to nest them to an arbitrary depth. So for example if you were creating a treasure table,
 * you might create a nested list such as:
 * MASTER_LIST
 *     COMMON_ITEMS
 *     MAGIC_ITEMS
 *         WEAPONS
 *             SWORDS
 *
 */
class RandomListItem {

  static final RandomListItem NULL_LIST_ITEM = new RandomListItem( "", null, null );

  /// the default weight for an item is a number that neither increases or decreases based on picks
  static const int WEIGHT_TYPE_DEFAULT = 0;

  /// the numeric weight decreases every time this item is picked.
  static const int WEIGHT_TYPE_DECREASES = 1;

  /// indicates the numeric weight increases everytime this item is NOT picked, but something else is
  static const int WEIGHT_TYPE_INCREASES = 2;

  /// item uses functions to calculate weight
  static const int WEIGHT_TYPE_DYNAMIC = 3;

  /// BOTH drop and increase specified
  static const int WEIGHT_TYPE_DECREASE_INCREASE = 4;


  String _id = "DEFAULT";
  /**
   * A unique case-insensitive identifier for this item. It should be unique across all items
   * so that queries into a list have no conflicts when trying to retrieve it.
   */
  String get id => _id;

  int _lifetime = -1;
  /**
   *  how many times item can be selected from list before being removed.
   *  default is -1, meaning infinite re-use
   */
  set lifetime(int lt ) => _lifetime = lt ;
  int get lifetime => _lifetime;


  /**
   * If the item weight decreases when picked, it will reset to this weight
   * when it drops below zero. By default this is set to the items original weight.
   *
   * If [removeOnZero] is true, the item instead will be removed from the list if the weight drops
   * to zero or below.
   */
  int maxWeight;

  RandomList get parentList => _parentList;
  /// The [RandomList] containing this item
  setParentList( RandomList p) {
    _parentList = p;
  }
  RandomList _parentList;

  /**
   *  The likelihood this item has of being picked from this list.
   *  Read-only.
   */
  double percent = 0.0; // for internal calculations

  /**
   *  Whether to remove item from list if its weight goes to zero or below.
   */
  bool removeOnZero = false;

  /// for internal calculations, public so that helper class can access it
  double threshold;

  /**
   * arbitrary tags applied to item
   */
  Map<String, Object> _tagList = new Map<String, Object>();

  /**
   *  How many times this item has been picked as a list result?
   */
  int pickCount = 0;

  /// Arbitrary data a system might want to store about an item. One
  /// use case would be to put a timestamp in here if that was important
  /// or script conditions
  Object userData;

  ///
  /// The numerical weight of the item if not calced by a weight function.
  int _weight = 0;

  /**
   * An arbitrary optional object that is used somehow to calculate the wt of the item.
   * Can be set externally and will be called as arg to the weight function.
   */
  Object weightData;

  int _weightDropAmount = 0;
  int get weightDropAmount => _weightDropAmount;

  /// not allowed below zero
  /// if greater than zero, weightDrops flag set to true, false otherwise
  void set weightDropAmount(int amount) {
    _weightDropAmount = (amount > -1) ? amount : 0;
    weightDrops = (amount > 0);
  }

  /**
   *  Whether the weight of the item drops when it is chosen by a list. Default: false
   */
  bool weightDrops = false;

  WeightFunction _weightFunction;

  ///  a function that is called with the weightData as an arg, and a callback for
  /// when the wt calculation is complete.
  WeightFunction get weightFunction => _weightFunction;

  /**
   *  A function that calcs the weight of the item in an arbitrary way.
   *  Imagine Thor's hammer: he can pick it up with ease, but to others it is
   *  infinitely heavy. A weight function could determine who is trying to pick it up.
   *
   *
   *  The [WeightFunction] gets passed the weightData and a callback as args.  The function must callback
   *  with an integer representing the calculated wt.
   */
  void set weightFunction(WeightFunction func) {
    if( weightType != WEIGHT_TYPE_DEFAULT )
      throw new Exception("Item " + id + " has a weightType conflict: trying to set a dynamic weight function when a drop or increase has already been set.");
    weightType = WEIGHT_TYPE_DYNAMIC;
    _weightFunction = func;
  }

  /**
   *  Every time item is NOT picked -- wt goes up by this amount.
   */
  int weightIncreaseAmount = 0;

  /// gets assigned one of the WEIGHT_TYPE const  - should be an enum
  int weightType = -1;


  /**
   *  An optional object that is 'wrapped' by this item.
   *
   *  It can be as simple as a string to display, it can be a completely functioning arbitrary game entity,
   *  a scripting langauge script to run, or any object that is conventient.
   *
   *  A special case is when the payload itself is a [RandomList]. In that case if this item is picked,
   *  the system will then make a pick from the child list and return the result of that.
   */
  Object payload;


  /// set an optional descriptive tag, with an option strength (value)
  setTag(String t, [Object value = 100]) {
    _tagList[ t ] = value;
  }

  removeTag(String t) {
    _tagList.remove(t);
  }

  bool hasTag(String t) => _tagList.containsKey(t);


  /////////////////////////////////////////////////////////////////////
  /**
   *
   */
  /// Default to infinite lifetime.
  /// @param id   auto-uppercased
  ///
  /// maxWeight will be set to the initial weight by default.
  /// @param wobject the payload of this item, if any.  Could be a function handler or entity, etc.
  /// @param weightDrops indicates whether the weight drops every time it is picked. Default false
  RandomListItem(String __id, this.payload, this._parentList, [int _weight_ = 100,
      int weightDrop = 0, int weightIncrease = 0, Object userData = null] )   {

    _id = __id.toUpperCase();

    _weight = _weight_;
    maxWeight = _weight_;
    weightType = WEIGHT_TYPE_DEFAULT;

    weightDropAmount = weightDrop;
    if( weightDrop > 0 ) weightType = WEIGHT_TYPE_DECREASES;

    if( weightIncrease > 0 ) {
      if( weightType == WEIGHT_TYPE_DECREASES )
        weightType  = WEIGHT_TYPE_DECREASE_INCREASE; // both types
      else {
        this.weightIncreaseAmount = weightIncrease;
        weightType = WEIGHT_TYPE_INCREASES;
      }
    }

    this.userData = userData;
  }


  /**
   *  add to the basic weight
   * [delta] can be negative
   *
   * throws exception if item has a weight function
   */
  addWeight( int delta ) {
    if( weightFunction != null ) {  // dynamic
      throw new Exception("Can't change the weight of an item with a weight function" + toString());
    }

    _weight += delta;
  }

  /// deep copy this item and any children
  RandomListItem clone() {
    Object newObj; // new wrapped object
    if( payload is RandomList ) {
      newObj = (payload as RandomList).clone(id);
    } else {
      newObj = payload;
    }
    RandomListItem item = new RandomListItem(id, newObj, _parentList );
    item.lifetime = lifetime;
    item.weight =  _weight ;
    item.threshold = threshold;
    item.percent = percent;
    item.weightDropAmount = weightDropAmount;
    item.weightIncreaseAmount = weightIncreaseAmount;
    item.weightData = weightData;
    item.weightFunction = weightFunction;
    return item;
  }


  /**
   *  this will be completed when a weight value is ready like so:
   *  _weightComplete.complete( value );
   */
  Completer _weightCompleter;



  /**
   * Calculate the current weight value of this item.
   *
   * returns either an integer if its a simple weight, or
   * a Future if it does a weight calc
   */
  Object get weight {
     if( weightFunction != null ) {
       _weightCompleter = new Completer();

       // call weight func and return val async
       weightFunction( this, weightData, (value){ _weightCompleter.complete(value); } );

       return _weightCompleter.future; // return completer
     } else {
       return _weight;  // simple
     }
  }




  /**
   * set the numeric weight of this item. The units are arbitrary and only relevant to each other.
   *
   * throws exception if the item has a weight function specified.
   */
  set weight( int newWeight ) {
    if( weightFunction != null ) {
      throw new Exception("Can't change the weight of an item with a weight function" + toString());
    }

    _weight = newWeight;
  }

  String toString() => id;

} // RandomListItem


